package models

import (
	bnet "blackforestbytes.com/badennet"
	"bytes"
	"encoding/base64"
	"encoding/json"
	"fmt"
	"gogs.mikescher.com/BlackForestBytes/goext/exerr"
	"gogs.mikescher.com/BlackForestBytes/goext/rext"
	"gogs.mikescher.com/BlackForestBytes/goext/rfctime"
	"golang.org/x/image/tiff"
	"golang.org/x/image/webp"
	"image"
	"image/gif"
	"image/jpeg"
	"image/png"
	"io"
	"os"
	"path"
	"regexp"
	"strings"
)

type BlobType string //@enum:type

const (
	BlobTypeOrgAttachment   BlobType = "ORG_ATTACHMENT"
	BlobTypeOrgAvatar       BlobType = "ORG_AVATAR"
	BlobTypeOrgBackground   BlobType = "ORG_BACKGROUND"
	BlobTypeOrgUserAvatar   BlobType = "ORGUSER_AVATAR"
	BlobTypePostImage       BlobType = "POST_IMAGE"
	BlobTypePostFile        BlobType = "POST_FILE"
	BlobTypeGroupAvatar     BlobType = "GROUP_AVATAR"
	BlobTypeGroupBackground BlobType = "GROUP_BACKGROUND"
	BlobTypeEventBackground BlobType = "EVENT_BACKGROUND"
)

var rexNonFilename = rext.W(regexp.MustCompile("[^A-Z0-9a-z_\\-ÄÖÜäöüß()\\[\\].,]"))

type Blob struct {
	BlobID       BlobID                   `json:"id"           bson:"_id,omitempty"`
	CreationTime rfctime.RFC3339NanoTime  `json:"creationTime" bson:"creationTime"`
	UpdateTime   rfctime.RFC3339NanoTime  `json:"updateTime"   bson:"updateTime"`
	CreatorType  EntryCreatorType         `json:"creatorType"  bson:"creatorType"`
	Creator      *AnyID                   `json:"creator"      bson:"creator"`
	Checksum     string                   `json:"checksum"     bson:"checksum"`
	Filesize     Bytes                    `json:"filesize"     bson:"filesize"`
	MimeType     MimeType                 `json:"mimeType"     bson:"mimeType"`
	Type         BlobType                 `json:"type"         bson:"type"`
	Reference    *AnyID                   `json:"reference"    bson:"reference"`
	Data         BlobData                 `json:"data"         bson:",inline"`
	Deleted      *rfctime.RFC3339NanoTime `json:"deleted"      bson:"deleted"`
}

func (b Blob) DownloadFilename() string {

	if b.Data.DataOrgAttachment != nil {
		if b.Data.DataOrgAttachment.OriginalFilename != "" {
			return b.Data.DataOrgAttachment.OriginalFilename
		}
		if fn := strings.TrimSpace(b.Data.DataOrgAttachment.Filename); fn != "" {
			fn = rexNonFilename.ReplaceAll(fn, "_", true)
			return fn + "." + b.MimeType.Extension()
		}
	}

	if b.Type == BlobTypeOrgAvatar {
		return "avatar." + b.MimeType.Extension()
	}

	if b.Type == BlobTypeOrgBackground {
		return "background." + b.MimeType.Extension()
	}

	if b.Type == BlobTypeGroupAvatar {
		return "avatar." + b.MimeType.Extension()
	}

	if b.Type == BlobTypeGroupBackground {
		return "background." + b.MimeType.Extension()
	}

	if b.Type == BlobTypeOrgUserAvatar {
		return "avatar." + b.MimeType.Extension()
	}

	return b.BlobID.String() + "." + b.MimeType.Extension()
}

func (b Blob) Filepath() string {
	return path.Join(bnet.Conf.DataDir, "blobs", b.BlobID.String())
}

func (b Blob) Read() ([]byte, error) {
	return os.ReadFile(b.Filepath())
}

func (b Blob) ReadBase64() (string, error) {
	fp := b.Filepath()

	f, err := os.Open(fp)
	if err != nil {
		return "", exerr.Wrap(err, "cannot open blob file").Id("blob.id", b.BlobID).Str("filepath", fp).Build()
	}

	buffer := &bytes.Buffer{}
	wc := base64.NewEncoder(base64.StdEncoding, buffer)

	_, err = io.Copy(wc, f)
	if err != nil {
		_ = wc.Close()
		return "", exerr.Wrap(err, "failed to copy blob inputstream to bytes.Buffer").Build()
	}

	err = wc.Close()
	if err != nil {
		return "", exerr.Wrap(err, "failed to close blob inputstream").Build()
	}

	return string(buffer.Bytes()), nil
}

func (b Blob) ReadImage() (image.Image, error) {
	fp := b.Filepath()

	r, err := os.Open(b.Filepath())
	if err != nil {
		return nil, exerr.Wrap(err, "cannot open blob file").Id("blob.id", b.BlobID).Str("filepath", fp).Build()
	}
	defer func() {
		err = r.Close()
		exerr.Wrap(err, "failed to close blob inputstream (in defer)").Print()
	}()

	if b.MimeType == "image/jpeg" {
		return jpeg.Decode(r)
	}

	if b.MimeType == "image/png" {
		return png.Decode(r)
	}

	if b.MimeType == "image/gif" {
		return gif.Decode(r)
	}

	if b.MimeType == "image/tiff" {
		return tiff.Decode(r)
	}

	if b.MimeType == "image/webp" {
		return webp.Decode(r)
	}

	return nil, exerr.New(bnet.ErrInvalidMimeType, fmt.Sprintf("cannot decode image of mimetype '%s'", b.MimeType)).Build()
}

type BlobData struct {
	DataOrgAttachment   *DataOrgAttachment   `bson:"dataOrgAttachment"`
	DataOrgAvatar       *DataOrgAvatar       `bson:"dataOrgAvatar"`
	DataOrgBackground   *DataOrgBackground   `bson:"dataOrgBackground"`
	DataOrgUserAvatar   *DataOrgUserAvatar   `bson:"dataOrgUserAvatar"`
	DataPostImage       *DataPostImage       `bson:"dataPostImage"`
	DataPostFile        *DataPostFile        `bson:"dataPostFile"`
	DataGroupAvatar     *DataGroupAvatar     `bson:"dataGroupAvatar"`
	DataGroupBackground *DataGroupBackground `bson:"dataGroupBackground"`
	DataEventBackground *DataEventBackground `bson:"dataEventBackground"`
}

func (d BlobData) MarshalJSON() ([]byte, error) {
	if d.DataOrgAttachment != nil {
		return json.Marshal(d.DataOrgAttachment)
	}
	if d.DataOrgAvatar != nil {
		return json.Marshal(d.DataOrgAvatar)
	}
	if d.DataOrgBackground != nil {
		return json.Marshal(d.DataOrgBackground)
	}
	if d.DataOrgUserAvatar != nil {
		return json.Marshal(d.DataOrgUserAvatar)
	}
	if d.DataPostImage != nil {
		return json.Marshal(d.DataPostImage)
	}
	if d.DataGroupAvatar != nil {
		return json.Marshal(d.DataGroupAvatar)
	}
	if d.DataGroupBackground != nil {
		return json.Marshal(d.DataGroupBackground)
	}
	return json.Marshal(nil)
}

func (d *BlobData) UnmarshalJSON(data []byte) error {
	return exerr.New(bnet.ErrUnmarshalJSON, "cannot unmarshal BlobData").Build()
}
